@@ -4,6 +4,9 @@
 #include "pl_util.h"
 #include "pl_console.h"
 
+#define NEED_newRV_noinc_GLOBAL
+#include "ppport.h"
+
 #define CONSOLE_FLUSH         0x01
 #define CONSOLE_TARGET_STDOUT 0x10
 #define CONSOLE_TARGET_STDERR 0x20
@@ -76,31 +79,45 @@ static int console_output(const FunctionCallbackInfo<Value>& args, unsigned int
     SV* message = newSVpvs("");
     bool separate = false;
     if (preamble) {
-        Perl_sv_catpv(aTHX_ message, preamble);
-        Perl_sv_catpvf(aTHX_ message, ":");
+        sv_catpv(aTHX_ message, preamble);
+        sv_catpvf(aTHX_ message, ":");
         separate = true;
     }
-    for (int j = start; j < args.Length(); j++) {
-        // add separator if necessary
-        if (separate) {
-            Perl_sv_catpvf(aTHX_ message, " ");
+
+    Local<Object> global = context->Global();
+    Local<Name> to_str_nam = String::NewFromUtf8(isolate, "JSON_stringify_with_cycles", NewStringType::kNormal).ToLocalChecked();
+    Local<Value> to_str_val = global->Get(to_str_nam);
+    if (to_str_val->IsFunction()) {
+        Local<Function> to_str_fun = Local<Function>::Cast(to_str_val);
+        Local<Value> sargs[1];
+        for (int j = start; j < args.Length(); j++) {
+            /* add separator if necessary */
+            if (separate) {
+                sv_catpvf(aTHX_ message, " ");
+            }
+            separate = true;
+
+            /* for non-objects, just get their value as string */
+            if (!args[j]->IsObject()) {
+                String::Utf8Value str(isolate, args[j]);
+                sv_catpvf(aTHX_ message, "%s", *str);
+                continue;
+            }
+
+            /* convert each object arg to JSON */
+            sargs[0] = args[j];
+            Local<Value> ret;
+            if (!to_str_fun->Call(context, global, 1, sargs).ToLocal(&ret)) {
+                continue;
+            }
+            Local<String> json = Local<String>::Cast(ret);
+            String::Utf8Value str(isolate, json);
+            sv_catpvf(aTHX_ message, "%s", *str);
         }
-        separate = true;
-        // for non-objects, just get their value as string
-        if (!args[j]->IsObject()) {
-            String::Utf8Value str(isolate, args[j]);
-            Perl_sv_catpvf(aTHX_ message, "%s", *str);
-            continue;
-        }
-        // convert objects to JSON
-        Local<String> json = JSON::Stringify(context, args[j]).ToLocalChecked();
-        String::Utf8Value str(isolate, json);
-        Perl_sv_catpvf(aTHX_ message, "%s", *str);
     }
-    Perl_sv_catpvf(aTHX_ message, "\n");
     if (stack) {
 #if 0
-        // TODO: generate a stack trace
+        /* TODO: generate a stack trace */
         v8_inspector::V8Inspector* inspector = new v8_inspector::V8InspectorImpl::V8InspectorImpl();
         inspector->captureStackTrace();
 #endif
@@ -203,7 +220,7 @@ int pl_show_error(V8Context* ctx, const char* fmt, ...)
     SV* message = newSVpvs("");
     va_list ap;
     va_start(ap, fmt);
-    Perl_sv_vcatpvf(aTHX_ message, fmt, &ap);
+    sv_vcatpvf(aTHX_ message, fmt, &ap);
     va_end(ap);
     return console_output_line(ctx, message, CONSOLE_TARGET_STDERR | CONSOLE_FLUSH);
 }
